# Copyright (C) 2020 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os

from swh.fuse.tests.data.config import REGULAR_FILE
from swh.model.identifiers import parse_swhid


def test_cache_artifact(fuse_mntdir):
    assert os.listdir(fuse_mntdir / "cache") == ["origin"]

    (fuse_mntdir / "archive" / REGULAR_FILE).is_file()

    swhid = parse_swhid(REGULAR_FILE)
    assert os.listdir(fuse_mntdir / "cache") == [swhid.object_id[:2], "origin"]


def test_purge_artifact(fuse_mntdir):
    DEFAULT_CACHE_CONTENT = ["origin"]

    assert os.listdir(fuse_mntdir / "cache") == DEFAULT_CACHE_CONTENT

    # Access a content artifact...
    (fuse_mntdir / "archive" / REGULAR_FILE).is_file()
    assert os.listdir(fuse_mntdir / "cache") != DEFAULT_CACHE_CONTENT
    # ... and remove it from cache
    swhid = parse_swhid(REGULAR_FILE)
    os.unlink(fuse_mntdir / "cache" / swhid.object_id[:2] / str(swhid))

    assert os.listdir(fuse_mntdir / "cache") == DEFAULT_CACHE_CONTENT



import os
from pathlib import Path

from swh.fuse.tests.common import get_data_from_web_archive
from swh.fuse.tests.data.config import FAKE_SNP_SPECIAL_CASES_SWHID, ROOT_SNP


def test_list_branches(fuse_mntdir):
    snp_dir = Path(fuse_mntdir / "archive" / ROOT_SNP)
    snp_meta = get_data_from_web_archive(ROOT_SNP)
    for branch_name in snp_meta.keys():
        assert (snp_dir / branch_name).is_symlink()


def test_special_cases(fuse_mntdir):
    snp_dir = Path(fuse_mntdir / "archive" / FAKE_SNP_SPECIAL_CASES_SWHID)
    snp_meta = get_data_from_web_archive(FAKE_SNP_SPECIAL_CASES_SWHID)
    for branch_name, branch_meta in snp_meta.items():
        curr = snp_dir / branch_name
        assert curr.is_symlink()
        if "expected_symlink" in branch_meta:
            assert os.readlink(curr) == branch_meta["expected_symlink"]


def test_access_rev_target(fuse_mntdir):
    dir_path = fuse_mntdir / "archive" / ROOT_SNP / "refs/heads/master"
    expected = set(["meta.json", "root", "parent", "parents", "history"])
    actual = set(os.listdir(dir_path))
    assert expected.issubset(actual)



#!/usr/bin/env python3

# Copyright (C) 2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


import json
from typing import Any, Dict

import requests

from swh.fuse.tests.api_url import (
    GRAPH_API_REQUEST,
    swhid_to_graph_url,
    swhid_to_web_url,
)
from swh.fuse.tests.data.config import (
    ALL_ENTRIES,
    FAKE_SNP_SPECIAL_CASES,
    FAKE_SNP_SPECIAL_CASES_SWHID,
    ORIGIN_URL,
    REV_SMALL_HISTORY,
)
from swh.model.identifiers import (
    CONTENT,
    DIRECTORY,
    RELEASE,
    REVISION,
    SNAPSHOT,
    SWHID,
    parse_swhid,
)

API_URL_real = "https://archive.softwareheritage.org/api/1"
API_URL_test = "https://invalid-test-only.archive.softwareheritage.org/api/1"

# Use your own API token to lift rate limiting. Note: this is not necessary to generate
# the API data only once but can be useful when re-generating it multiple times.
API_TOKEN = ""


MOCK_ARCHIVE: Dict[str, Any] = {}
# Temporary map (swhid -> metadata) to ease data generation
METADATA: Dict[SWHID, Any] = {}


def get_from_api(endpoint: str) -> str:
    headers = {"Authorization": f"Bearer {API_TOKEN}"} if API_TOKEN else {}
    return requests.get(f"{API_URL_real}/{endpoint}", headers=headers).text


def get_short_type(object_type: str) -> str:
    short_type = {
        CONTENT: "cnt",
        DIRECTORY: "dir",
        REVISION: "rev",
        RELEASE: "rel",
        SNAPSHOT: "snp",
    }
    return short_type[object_type]


def generate_archive_web_api(
    swhid: SWHID, raw: bool = False, recursive: bool = False
) -> None:
    # Already in mock archive
    if swhid in METADATA and not raw:
        return

    url = swhid_to_web_url(swhid, raw)

    data = get_from_api(url)
    if not raw:
        data = json.loads(data)

    MOCK_ARCHIVE[url] = data
    METADATA[swhid] = data

    # Retrieve additional needed data for different artifacts (eg: content's
    # blob data, release target, etc.)
    if recursive:
        if swhid.object_type == CONTENT:
            generate_archive_web_api(swhid, raw=True)
        elif swhid.object_type == RELEASE:
            target_type = METADATA[swhid]["target_type"]
            target_id = METADATA[swhid]["target"]
            target = parse_swhid(f"swh:1:{get_short_type(target_type)}:{target_id}")
            generate_archive_web_api(target, recursive=True)


def generate_archive_graph_api(swhid: SWHID) -> None:
    if swhid.object_type == REVISION:
        # Empty history for all revisions (except REV_SMALL_HISTORY used in tests)
        url = swhid_to_graph_url(swhid, GRAPH_API_REQUEST.HISTORY)
        MOCK_ARCHIVE[url] = ""
        if str(swhid) == REV_SMALL_HISTORY:
            # TODO: temporary fix, retrieve from the graph API once it is public
            history = """
swh:1:rev:37426e42cf78a43779312d780eecb21a64006d99 swh:1:rev:0cf3c2ad935be699281ed20fb3d2f29554e6229b
swh:1:rev:0cf3c2ad935be699281ed20fb3d2f29554e6229b swh:1:rev:37180552769b316e7239d047008f187127e630e6
swh:1:rev:37180552769b316e7239d047008f187127e630e6 swh:1:rev:dd2716f56c7cf55f2904fbbf4dfabaab1afbcd88
swh:1:rev:dd2716f56c7cf55f2904fbbf4dfabaab1afbcd88 swh:1:rev:968ec145278d3d6562e4b5ec4006af97dc0da563
swh:1:rev:968ec145278d3d6562e4b5ec4006af97dc0da563 swh:1:rev:34dc7053ebfd440648f49dc83d2538ab5e7ceda5
swh:1:rev:34dc7053ebfd440648f49dc83d2538ab5e7ceda5 swh:1:rev:c56a729ff1d9467d612bf522614519ac7b97f798
swh:1:rev:c56a729ff1d9467d612bf522614519ac7b97f798 swh:1:rev:eb7807c4fe7a2c2ad3c074705fb70de5eae5abe3
swh:1:rev:eb7807c4fe7a2c2ad3c074705fb70de5eae5abe3 swh:1:rev:d601b357ecbb1fa33dc10c177bb557868be07deb
swh:1:rev:d601b357ecbb1fa33dc10c177bb557868be07deb swh:1:rev:2a2474d497ae19472b4366f6d8d62e9a516787c3
swh:1:rev:2a2474d497ae19472b4366f6d8d62e9a516787c3 swh:1:rev:eed5c0aa249f3e17bbabeeba1650ab699e3dff5a
swh:1:rev:eed5c0aa249f3e17bbabeeba1650ab699e3dff5a swh:1:rev:67d1f0a9aafaa7dcd63b86032127ab660e630c46
swh:1:rev:67d1f0a9aafaa7dcd63b86032127ab660e630c46 swh:1:rev:2e3fa5bd68677762c619d83dfdf1a83ba7f0e749
swh:1:rev:2e3fa5bd68677762c619d83dfdf1a83ba7f0e749 swh:1:rev:a9c639ec8af3a4099108788c1db0176c7fea5799
swh:1:rev:a9c639ec8af3a4099108788c1db0176c7fea5799 swh:1:rev:c06ea8f9445dbb5eda99ac8730d7fb2177df6816
swh:1:rev:c06ea8f9445dbb5eda99ac8730d7fb2177df6816 swh:1:rev:422b8a6be4aab120685f450db0a520fcb5a8aa6b
swh:1:rev:422b8a6be4aab120685f450db0a520fcb5a8aa6b swh:1:rev:e8759934711c70c50b5d616be22104e649abff58
swh:1:rev:e8759934711c70c50b5d616be22104e649abff58 swh:1:rev:63b5e18207c7f8a261c1f7f50fd8c7bbf9a21bda
swh:1:rev:63b5e18207c7f8a261c1f7f50fd8c7bbf9a21bda swh:1:rev:5dfe101e5197d6854aa1d8c9907ac7851468d468
swh:1:rev:5dfe101e5197d6854aa1d8c9907ac7851468d468 swh:1:rev:287d69ddacba3f5945b70695fb721b2f055d3ee6
swh:1:rev:287d69ddacba3f5945b70695fb721b2f055d3ee6 swh:1:rev:85a701c8f668fc03e6340682956e7ca7d9cf54bc
swh:1:rev:85a701c8f668fc03e6340682956e7ca7d9cf54bc swh:1:rev:241305caab232b04666704dc6853c41312cd283a
swh:1:rev:241305caab232b04666704dc6853c41312cd283a swh:1:rev:0d9565a4c144c07dab052161eb5fa3815dcd7f06
swh:1:rev:0d9565a4c144c07dab052161eb5fa3815dcd7f06 swh:1:rev:72c6c60d80cdfe63af5046a1a98549f0515734f2
swh:1:rev:72c6c60d80cdfe63af5046a1a98549f0515734f2 swh:1:rev:c483808e0ff9836bc1cda0ce95d77c8b7d3be91c
swh:1:rev:c483808e0ff9836bc1cda0ce95d77c8b7d3be91c swh:1:rev:1c60be2f32f70f9181a261ae2c2b4efe353d0f85
swh:1:rev:1c60be2f32f70f9181a261ae2c2b4efe353d0f85 swh:1:rev:bcf29b882acdf477be412fdb401b0fc2a6c819aa
swh:1:rev:bcf29b882acdf477be412fdb401b0fc2a6c819aa swh:1:rev:261d543920e1c66049c469773ca989aaf9ce480e
swh:1:rev:261d543920e1c66049c469773ca989aaf9ce480e swh:1:rev:24d5ff75c3abfe7b327c48468ed9a39f0d8a0427
swh:1:rev:24d5ff75c3abfe7b327c48468ed9a39f0d8a0427 swh:1:rev:d3c0762ff85ff7d29668d1f5d2361df03978bbea
swh:1:rev:d3c0762ff85ff7d29668d1f5d2361df03978bbea swh:1:rev:af44ec2856603b8a978a1f2582c285c7c0065403
swh:1:rev:af44ec2856603b8a978a1f2582c285c7c0065403 swh:1:rev:69a34503f4d51b639855501f1b6d6ce2da4e16c7
swh:1:rev:69a34503f4d51b639855501f1b6d6ce2da4e16c7 swh:1:rev:0364a801bb29211d4731f3f910c7629286b51c45
swh:1:rev:0364a801bb29211d4731f3f910c7629286b51c45 swh:1:rev:25eb1fd3c9d997e460dff3e03d87e398e616c726
swh:1:rev:25eb1fd3c9d997e460dff3e03d87e398e616c726 swh:1:rev:4a1f86ccd7e823f63d12208baef79b1e74479203
swh:1:rev:4a1f86ccd7e823f63d12208baef79b1e74479203 swh:1:rev:0016473117e4bc3c8959bf2fd49368844847d74c
swh:1:rev:0016473117e4bc3c8959bf2fd49368844847d74c swh:1:rev:935442babcf4f8ae52c1a13bb9ce07270a302886
swh:1:rev:935442babcf4f8ae52c1a13bb9ce07270a302886 swh:1:rev:1f3cff91f6762b0f47f41025b5e2c5ac942479ba
swh:1:rev:1f3cff91f6762b0f47f41025b5e2c5ac942479ba swh:1:rev:bc286c7f2ceb5c3d2e06ec72f78d28842f94ef65
swh:1:rev:bc286c7f2ceb5c3d2e06ec72f78d28842f94ef65 swh:1:rev:f038f4d533f897a29f9422510d1b3f0caac97388
swh:1:rev:f038f4d533f897a29f9422510d1b3f0caac97388 swh:1:rev:d6b7c96c3eb29b9244ece0c046d3f372ff432d04
swh:1:rev:d6b7c96c3eb29b9244ece0c046d3f372ff432d04 swh:1:rev:c01efc669f09508b55eced32d3c88702578a7c3e
"""  # NoQA: E501
            MOCK_ARCHIVE[url] = history

            hist_nodes = set(
                map(
                    parse_swhid,
                    [edge.split(" ")[1] for edge in history.strip().split("\n")],
                )
            )
            for swhid in hist_nodes:
                generate_archive_web_api(swhid, recursive=False)


def generate_origin_archive_web_api(url: str):
    url_visits = f"origin/{url}/visits/"
    data = get_from_api(url_visits)
    data = json.loads(data)
    MOCK_ARCHIVE[url_visits] = data
    # Necessary since swh-fuse will check the origin URL using the get/ endpoint
    url_get = f"origin/{url}/get/"
    MOCK_ARCHIVE[url_get] = ""


for entry in ALL_ENTRIES:
    swhid = parse_swhid(entry)
    generate_archive_web_api(swhid, recursive=True)
    generate_archive_graph_api(swhid)

# Custom fake snapshot to handle most special cases
MOCK_ARCHIVE[swhid_to_web_url(FAKE_SNP_SPECIAL_CASES_SWHID)] = {
    "branches": FAKE_SNP_SPECIAL_CASES
}

# Origin artifacts are not identified by SWHID but using an URL
generate_origin_archive_web_api(ORIGIN_URL)

print("# GENERATED FILE, DO NOT EDIT.")
print("# Run './gen-api-data.py > api_data.py' instead.")
print("# flake8: noqa")
print("from typing import Any, Dict")
print("")
print(f"API_URL = '{API_URL_test}'\n")
print(f"MOCK_ARCHIVE: Dict[str, Any] = {MOCK_ARCHIVE}")



# Copyright (C) 2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

# WARNING: do not import unnecessary things here to keep cli startup time under
# control
import os
from pathlib import Path
from typing import Any, Dict

import click

from swh.core.cli import CONTEXT_SETTINGS
from swh.core.cli import swh as swh_cli_group
from swh.model.cli import SWHIDParamType

# All generic config code should reside in swh.core.config
DEFAULT_CONFIG_PATH = os.environ.get(
    "SWH_CONFIG_FILE", os.path.join(click.get_app_dir("swh"), "global.yml")
)

CACHE_HOME_DIR: Path = (
    Path(os.environ["XDG_CACHE_HOME"])
    if "XDG_CACHE_HOME" in os.environ
    else Path.home() / ".cache"
)

DEFAULT_CONFIG: Dict[str, Any] = {
    "cache": {
        "metadata": {"path": str(CACHE_HOME_DIR / "swh/fuse/metadata.sqlite")},
        "blob": {"path": str(CACHE_HOME_DIR / "swh/fuse/blob.sqlite")},
        "direntry": {"maxram": "10%"},
    },
    "web-api": {
        "url": "https://archive.softwareheritage.org/api/1",
        "auth-token": None,
    },
    "json-indent": 2,
}


@swh_cli_group.group(name="fs", context_settings=CONTEXT_SETTINGS)
@click.option(
    "-C",
    "--config-file",
    default=None,
    type=click.Path(exists=True, dir_okay=False, path_type=str),
    help=f"Configuration file (default: {DEFAULT_CONFIG_PATH})",
)
@click.pass_context
def fuse(ctx, config_file):
    """Software Heritage virtual file system"""

    import logging
    from shutil import which

    import yaml

    from swh.core import config

    if which("fusermount3") is None:
        logging.error("Missing dependency: 'fusermount3'")
        ctx.exit(1)

    if not config_file:
        config_file = DEFAULT_CONFIG_PATH

    try:
        conf = config.read_raw_config(config.config_basepath(config_file))
        if not conf:
            raise ValueError(f"Cannot parse configuration file: {config_file}")

        if config_file == DEFAULT_CONFIG_PATH:
            try:
                conf = conf["swh"]["fuse"]
            except KeyError:
                pass

        # recursive merge not done by config.read
        conf = config.merge_configs(DEFAULT_CONFIG, conf)
    except Exception:
        logging.warning(
            "Using default configuration (cannot load custom one)", exc_info=True
        )
        conf = DEFAULT_CONFIG

    logging.debug("Read configuration: \n%s", yaml.dump(conf))
    ctx.ensure_object(dict)
    ctx.obj["config"] = conf


@fuse.command(name="mount")
@click.argument(
    "path",
    required=True,
    metavar="PATH",
    type=click.Path(exists=True, dir_okay=True, file_okay=False),
)
@click.argument("swhids", nargs=-1, metavar="[SWHID]...", type=SWHIDParamType())
@click.option(
    "-f/-d",
    "--foreground/--daemon",
    default=False,
    help="whether to run FUSE attached to the console (foreground) "
    "or daemonized in the background (default: daemon)",
)
@click.pass_context
def mount(ctx, swhids, path, foreground):
    """Mount the Software Heritage virtual file system at PATH.

    If specified, objects referenced by the given SWHIDs will be prefetched and used to
    populate the virtual file system (VFS). Otherwise the VFS will be populated
    on-demand, when accessing its content.

    \b
    Example:

    \b
      $ mkdir swhfs
      $ swh fs mount swhfs/
      $ grep printf swhfs/archive/swh:1:cnt:c839dea9e8e6f0528b468214348fee8669b305b2
          printf("Hello, World!");
      $

    """

    import asyncio
    from contextlib import ExitStack
    import logging

    from daemon import DaemonContext

    from swh.fuse import LOGGER_NAME, fuse

    # TODO: set default logging settings when --log-config is not passed
    # DEFAULT_LOG_PATH = Path(".local/swh/fuse/mount.log")
    with ExitStack() as stack:
        if not foreground:
            # TODO: temporary fix until swh.core has the proper logging utilities
            # Disable logging config before daemonizing, and reset it once
            # daemonized to be sure to not close file handlers
            log_level = logging.getLogger(LOGGER_NAME).getEffectiveLevel()
            logging.shutdown()
            # Stay in the current working directory when spawning daemon
            cwd = os.getcwd()
            stack.enter_context(DaemonContext(working_directory=cwd))
            logging.config.dictConfig(
                {
                    "version": 1,
                    "handlers": {
                        "syslog": {
                            "class": "logging.handlers.SysLogHandler",
                            "address": "/dev/log",
                        },
                    },
                    "loggers": {
                        LOGGER_NAME: {"level": log_level, "handlers": ["syslog"],},
                    },
                }
            )

        conf = ctx.obj["config"]
        asyncio.run(fuse.main(swhids, path, conf))


@fuse.command()
@click.argument(
    "path",
    required=True,
    metavar="PATH",
    type=click.Path(exists=True, dir_okay=True, file_okay=False),
)
@click.pass_context
def umount(ctx, path):
    """Unmount a mounted virtual file system.

    Note: this is equivalent to ``fusermount -u PATH``, which can be used to unmount any
    FUSE-based virtual file system. See ``man fusermount3``.

    """
    import logging
    import subprocess

    try:
        subprocess.run(["fusermount", "-u", path], check=True)
    except subprocess.CalledProcessError as err:
        logging.error(
            "cannot unmount virtual file system: '%s' returned exit status %d",
            " ".join(err.cmd),
            err.returncode,
        )
        ctx.exit(1)


@fuse.command()
@click.pass_context
def clean(ctx):
    """Clean on-disk cache(s).

    """

    def rm_cache(conf, cache_name):
        try:
            Path(conf["cache"][cache_name]["path"]).unlink()
        except (FileNotFoundError, KeyError):
            pass

    conf = ctx.obj["config"]
    for cache_name in ["blob", "metadata"]:
        rm_cache(conf, cache_name)
